import secrets
import datetime

from django.db import models
from django.contrib.auth.models import User
from django.utils.translation import gettext_lazy as _

from cryptography import x509
from cryptography.hazmat.primitives import serialization, hashes
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.x509.oid import NameOID
from cryptography.hazmat.primitives import hashes

VERIFICATION_PENDING = 0
VERIFICATION_SUCCESS = 10
VERIFICATION_FAILED = 20

VERIFICATION_STATUS = [
    (VERIFICATION_PENDING, "Verification of the CSR is pending"),
    (VERIFICATION_SUCCESS, "Verification succeeded"),
    (VERIFICATION_FAILED, "Verification failed"),
]

INFO_MAP = {
    "country": NameOID.COUNTRY_NAME,
    "state_or_province": NameOID.STATE_OR_PROVINCE_NAME,
    "locality": NameOID.LOCALITY_NAME,
    "organization": NameOID.ORGANIZATION_NAME,
    "common_name": NameOID.COMMON_NAME
}

class CertificationAuthority(models.Model):
    key_pem = models.TextField(null=True, blank=True)
    parent = models.ForeignKey(
        'CertificationAuthority',
        on_delete=models.PROTECT,
        related_name='children',
        null=True, blank=True
    )
    cert = models.ForeignKey(
        'Certificate',
        on_delete=models.SET_NULL,
        related_name=None,
        null=True, blank=True
    )

    comment = models.CharField(max_length=128)
    title = models.CharField(max_length=32)

    country = models.CharField(max_length=128, null=True, blank=True)
    state_or_province = models.CharField(max_length=128, null=True, blank=True)
    locality = models.CharField(max_length=128, null=True, blank=True)
    organization = models.CharField(max_length=128, null=True, blank=True)
    common_name = models.CharField(max_length=128, null=True, blank=True)

    created = models.DateTimeField(auto_now_add=True)

    class Meta:
        permissions = [
            (
                "manage_certificationauthority",
                "Can issue certificates signed by the CA",
            ),
            (
                "request_certificate",
                "Can request a certificate from the CA",
            )
        ]
            

    def generate_key(self, passphrase, length=4096):
        # Generate our key
        key = rsa.generate_private_key(
            public_exponent=65537,
            key_size=length,
        )

        self.key_pem = key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.TraditionalOpenSSL,
            encryption_algorithm=serialization.BestAvailableEncryption(passphrase)
        ).decode()


    def create_self_sign_request(self, password):
        """Returns CSR"""
        name = []
        for label, oid in INFO_MAP.items():
            value = getattr(self, label)
            if value:
                name.append(x509.NameAttribute(INFO_MAP[label], value))

        key = serialization.load_pem_private_key(
            self.key_pem.encode(),
            password,
        )

        csr = x509.CertificateSigningRequestBuilder().subject_name(
            x509.Name(name)
        ).add_extension(
            x509.BasicConstraints(ca=True, path_length=1),
            critical=True
        ).sign(key, hashes.SHA256())

        csr_obj = CertificateSigningRequest()
        csr_obj.ca = self
        csr_obj.pem = csr.public_bytes(serialization.Encoding.PEM).decode()
        csr_obj.save()
        return csr_obj
        
    def approve(self, csr_obj, password, days=0):
        """Returns certificate"""
        key = serialization.load_pem_private_key(
            self.key_pem.encode(),
            password,
        )

        name = []
        for label, oid in INFO_MAP.items():
            value = getattr(self, label)
            if value:
                name.append(x509.NameAttribute(INFO_MAP[label], value))

        serial_number = x509.random_serial_number()

        csr = x509.load_pem_x509_csr(
            csr_obj.pem.encode(),
        )

        cert_builder = x509.CertificateBuilder().subject_name(
            csr.subject
        ).issuer_name(
            x509.Name(name)
        ).public_key(
            csr.public_key()
        ).serial_number(
            serial_number
        ).not_valid_before(
            datetime.datetime.utcnow()
        ).not_valid_after(
            datetime.datetime.utcnow() + datetime.timedelta(days=days)
        )
        for extension in csr.extensions:
            cert_builder = cert_builder.add_extension(extension.value, critical=extension.critical)
        cert = cert_builder.sign(key, hashes.SHA256())

        cert_obj = Certificate()
        cert_obj.serial_number = serial_number.to_bytes(20, byteorder='big')
        cert_obj.pem = cert.public_bytes(serialization.Encoding.PEM).decode()
        cert_obj.request = csr_obj
        cert_obj.ca = self

        csr_obj.verification = VERIFICATION_SUCCESS
        csr_obj.save()

        cert_obj.save()
        return cert_obj

    def __str__(self):
        return self.title


class CertificateSigningRequest(models.Model):
    ca = models.ForeignKey(
        CertificationAuthority,
        on_delete=models.PROTECT,
        related_name='requests'
    )
    pem = models.TextField()

    pickup_code = models.CharField(
        max_length=64,
        default=secrets.token_urlsafe
    )
    created = models.DateTimeField(auto_now_add=True)

    verification = models.IntegerField(
        _("Verification status"),
        choices=VERIFICATION_STATUS,
        default=VERIFICATION_PENDING,
    )

    def verification_text(self):
        return list(filter(
            lambda x: self.verification == x[0],
            VERIFICATION_STATUS
        ))[0][1]

    def __str__(self):
        return f"{self.pk} ({self.verification_text()})"

    def __init__(self, *args, **kwds):
        super().__init__(*args, **kwds)
        self.decoded = None

    def _decode(self):
        """Decode the certificate and cache it internally"""
        if self.decoded:
            return True

        try:
            self.decoded = x509.load_pem_x509_csr(self.pem.encode())
        except ValueError:
            return False

        return True

    def subject(self):
        """Return the subject of the certificate"""
        if not self._decode():
            return None

        return self.decoded.subject.rfc4514_string()

    def extensions(self):
        """Return the subject of the certificate"""
        return [(
            e.oid._name,
            e.value,
            e.critical
        ) for e in self.decoded.extensions]

    def extensions_human(self):
        """Return the subject of the certificate"""
        if not self._decode():
            return None

        def remove_prefix(text, prefix):
            if text.startswith(prefix):
                return text[len(prefix):]
            return text

        def encode(d):
            pieces = [f"{remove_prefix(k, '_')}={v}" for k, v in vars(d).items()]
            return ", ".join(pieces)

        return [(
            v.__class__.__name__,
            encode(v),
            v
        ) for n, v, c in self.extensions()]

class Certificate(models.Model):
    ca = models.ForeignKey(
        CertificationAuthority,
        on_delete=models.PROTECT,
        related_name='issued_certificates'
    )
    request = models.ForeignKey(
        CertificateSigningRequest,
        on_delete=models.PROTECT,
        related_name='certificates'
    )
    pem = models.TextField()
    created = models.DateTimeField(auto_now_add=True)


    def __init__(self, *args, **kwds):
        super().__init__(*args, **kwds)
        self.decoded = None

    def _decode(self):
        """Decode the certificate and cache it internally"""
        if self.decoded:
            return True

        try:
            self.decoded = x509.load_pem_x509_certificate(self.pem.encode())
        except ValueError:
            return False

        return True

    def serial_number(self):
        """Return the subject of the certificate"""
        if not self._decode():
            return None

        return self.decoded.serial_number.to_bytes(20, 'big')

    def serial_number_human(self):
        """Return the subject of the certificate"""
        serial = self.serial_number().hex()
        return ":".join(a + b for a, b in zip(serial[::2], serial[1::2]))


    def subject(self):
        """Return the subject of the certificate"""
        if not self._decode():
            return None

        return self.decoded.subject.rfc4514_string()

    def issuer(self):
        """Return the issuer of the certificate"""
        if not self._decode():
            return None

        return self.decoded.issuer.rfc4514_string()

    def not_valid_before(self):
        """Return the utc representation of not valid before time"""
        if not self._decode():
            return None
        return self.decoded.not_valid_before

    def not_valid_after(self):
        """Return the utc representation of not valid after time"""
        if not self._decode():
            return None
        return self.decoded.not_valid_after

    def fingerprint(self):
        """Return the fingerprint of the certificate"""
        if not self._decode():
            return None
        return self.decoded.fingerprint(hashes.SHA256())

    def fingerprint_human(self):
        """Return the fingerprint of the certificate"""
        fingerprint = self.fingerprint().hex()
        return fingerprint

    def extensions(self):
        """Return the subject of the certificate"""
        return [(
            e.oid._name,
            e.value,
            e.critical
        ) for e in self.decoded.extensions]

    def extensions_human(self):
        """Return the subject of the certificate"""
        if not self._decode():
            return None

        def remove_prefix(text, prefix):
            if text.startswith(prefix):
                return text[len(prefix):]
            return text

        def encode(d):
            pieces = [f"{remove_prefix(k, '_')}={v}" for k, v in vars(d).items()]
            return ", ".join(pieces)

        return [(
            v.__class__.__name__,
            encode(v),
            v
        ) for n, v, c in self.extensions()]

    def __str__(self):
        return f"{self.ca}: {self.serial_number().hex()}"

